/************************************************************************
 *   IRC - Internet Relay Chat, src/ssl.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: ssl.c,v 1.14 2002/05/23 17:51:47 smoli Exp $
 */

#include "config.h"
#include "s_log.h"
#include "ssl.h"
#include "struct.h"
#include "common.h"
#include "client.h"
#include "s_log.h"
#include "s_bsd.h"
#include "send.h"
#include "client.h"
#include "numeric.h"
#include "ircd.h"
#include "s_user.h"
#include "send.h"

#ifdef HAVE_SSL

#define SSL_READ    1
#define SSL_WRITE   2
#define SSL_ACCEPT  3
#define SSL_CONNECT 4

#include <string.h>
#include <sys/stat.h>
#include <openssl/rand.h>
#include <openssl/asn1.h>

int         ssl_capable = 0;
int         ssl_cvl = 0;               /* client verification level */
int         ssl_svl = 0;               /* server verification level */
char       *ssl_cipher_list = NULL;
char       *ssl_ciphers = SSL_DEFAULT_CIPHER_LIST;
char       *ssl_pem_passphrase = NULL; /* is having the passphrase in ircd.conf a security leak? */
SSL_CTX    *server_ctx;
SSL_METHOD *server_meth;
SSL_CTX    *client_ctx;
SSL_METHOD *client_meth;

static int server_verify_peer = 0;
static int client_verify_peer = 0;

#define CTX_ERR(x) { \
                     ERR_print_errors_fp(stderr); \
                     SSL_CTX_free((x)); \
                     return 0; \
                   }

#define CHK_NULL(x) if(!x) \
                      { \
                        log(L_NOTICE, "Lost connection to %s: Error in SSL", \
                            get_client_name(cptr, TRUE)); \
                        return 0; \
                      }

int ssl_init();
static SSL_CTX *ssl_create_context(SSL_METHOD *method);
int ssl_init_verify(SSL_CTX *, int (*)(int, X509_STORE_CTX *), int, int *);

/* some forward declarations */
static time_t ASN1_UTCTIME_get(const ASN1_UTCTIME *s);
static char *make_cipher_string(const char *);
static int init_diffie_hellman();
static int ssl_server_verify(int, X509_STORE_CTX *);
static int ssl_client_verify(int, X509_STORE_CTX *);
static int ssl_verify(int, X509_STORE_CTX *, int, int);
static int ssl_error(int, int, struct Client *);
#if SSLEAY_VERSION_NUMBER >= 0x00904000L
static int ssl_pem_password_cb(char *, int, int, void *);
#else
static int ssl_pem_password_cb(char *, int, int);
#endif
static RSA *ssl_tmp_rsa_cb(SSL *, int, int);

#ifndef HAVE_ASN1_UTCTIME_GET /* current ssl versions (0.9.6) have commented this out */
static time_t ASN1_UTCTIME_get(const ASN1_UTCTIME *s)
{
  struct tm tm;
  int offset;
   
  memset(&tm,'\0',sizeof tm);
  
  #define g2(p) (((p)[0] - '0') * 10 + (p)[1] - '0')
  
  tm.tm_year = g2(s->data);
  if(tm.tm_year < 50)
    tm.tm_year += 100;
  
  tm.tm_mon = g2(s->data + 2)-1;
  tm.tm_mday = g2(s->data + 4);
  tm.tm_hour = g2(s->data + 6);
  tm.tm_min = g2(s->data + 8);
  tm.tm_sec = g2(s->data + 10);
  
  if(s->data[12] == 'Z')
    offset = 0;
  else
    {
      offset = g2(s->data + 13) * 60 + g2(s->data + 15);
      if(s->data[12] == '-')
        offset = -offset;
    }
  
#undef g2
  
  return mktime(&tm) - offset * 60; /* FIXME: mktime assumes the current timezone
                                     * instead of UTC, and unless we rewrite OpenSSL
                                     * in Lisp we cannot locally change the timezone
                                     * without possibly interfering with other parts
                                     * of the program. timegm, which uses UTC, is
                                     * non-standard.
                                     * Also time_t is inappropriate for general
                                     * UTC times because it may a 32 bit type. */
}
#endif

/* need different cipher delimiter (':' is already used as field delimiter)
 * 
 * replace '/' with ':'
 */
static char *make_cipher_string(const char *str)
{
  static char buf[1024];
  int x;
  
  for(x = 0; (x < 1023) && str[x]; x++)
    {
      if(str[x] != '/')
        buf[x] = str[x];
      else
        buf[x] = ':';
    }
  
  buf[x] = '\0';
  
  return buf;
}

/*
 * init client/server context and verification stuff.
 * this MUST be called AFTER parsing the config file,
 * because of S: line.
 */
int ssl_init()
{
  log(L_DEBUG, "SSL Line - passphrase: %s cvl: %u svl: %u",
      ssl_pem_passphrase, ssl_cvl, ssl_svl);
  
  SSL_load_error_strings();
  SSLeay_add_ssl_algorithms();
  
  server_meth = SSLv23_server_method();  
  client_meth = SSLv3_client_method();  
  
  if(ssl_cipher_list && ssl_cipher_list[0])
    ssl_ciphers = make_cipher_string(ssl_cipher_list);
  
  server_ctx = ssl_create_context(server_meth);
  client_ctx = ssl_create_context(client_meth);

  log(L_NOTICE, "SSL cipher list: %s", ssl_ciphers);
  
  if(!server_ctx || !client_ctx)
    return(log(L_ERROR, "Could not create SSL context") == NULL);
  
/*  if(init_diffie_hellman(server_ctx))
    log(L_WARN, "Couldn't load Diffie-Hellman parameters from %s", CEPATH);*/
  
  if(ssl_init_verify(server_ctx, ssl_server_verify, ssl_svl, &server_verify_peer) ||
     ssl_init_verify(client_ctx, ssl_client_verify, ssl_cvl, &client_verify_peer))
    log(L_WARN, "Could not init certificate verification");
  
  return 0;
}

/*
 * create a context, set the PEM passphrase callback, enable
 * session cache, set timeouts and load certificate.
 */
static SSL_CTX *ssl_create_context(SSL_METHOD *method)
{
  SSL_CTX *ret;
  
  ret = SSL_CTX_new(method);
  
  if(ret == NULL)
    return NULL;
  
  SSL_CTX_set_default_passwd_cb(ret, ssl_pem_password_cb);
  SSL_CTX_set_tmp_rsa_callback(ret, ssl_tmp_rsa_cb);
  
#if SSLEAY_VERSION_NUMBER >= 0x00906000L
  SSL_CTX_set_mode(ret, SSL_MODE_ENABLE_PARTIAL_WRITE |
                        SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
#endif  
  
  SSL_CTX_set_session_cache_mode(ret, SSL_SESS_CACHE_BOTH);
  SSL_CTX_set_timeout(server_ctx, 360);
  SSL_CTX_set_timeout(client_ctx, 360);
  
  /* init certificate stuff for context */
  if(!SSL_CTX_use_certificate_chain_file(ret, CEPATH))
    return log(L_ERROR, "Error using certificate chain file '%s'", CEPATH);
  
  if(!SSL_CTX_use_RSAPrivateKey_file(ret, CEPATH, SSL_FILETYPE_PEM))
    return log(L_ERROR, "Error using private key file '%s'", CEPATH);
    
  if(!SSL_CTX_check_private_key(ret))
    return log(L_ERROR, "Certificate and private key do not match");
  
  if(!SSL_CTX_set_cipher_list(ret, ssl_ciphers))
    return log(L_ERROR, "Error setting cipher list");
   
  return ret;
}

/*
 * setup certificate verification stuff (callbacks etc.)
 */
int ssl_init_verify(SSL_CTX *ctx, int (*verify_callback)(int, X509_STORE_CTX *), int verify_level, int *verify_peer)
{
  int verify = 0;

  /* init certificate verification stuff */
  if(!SSL_CTX_set_default_verify_paths(ctx))
    return(log(L_ERROR, "Could not set default verify paths") == NULL);
  
  if(!SSL_CTX_load_verify_locations(ctx, CAPATH, NULL))
    return(log(L_ERROR, "Error loading verify certificates from %s", CAPATH) == NULL);
  
  SSL_CTX_set_client_CA_list(ctx, SSL_load_client_CA_file(CAPATH));
  
  switch(verify_level)
    {
      case 3:
        *verify_peer = 1;
      case 2:
        verify |= SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
      case 1:
        verify |= SSL_VERIFY_PEER;
      case 0:
        break;
      default:
        log(L_WARN, "Bad verify level");
    }
  
  SSL_CTX_set_verify(ctx, verify ? verify : SSL_VERIFY_NONE, verify_callback);
  
/*  stack = SSL_CTX_get_client_CA_list(ctx);
  log(L_DEBUG, "There are %d CA_list things", sk_X509_NAME_num(stack));*/

  /* SSL_CTX_set_info_callback(server_ctx, ssl_info_callback);
     SSL_CTX_set_info_callback(client_ctx, ssl_info_callback); */
  return 0;
}

/*
 * just return the passphrase specified in the config file.
 * return error when no passphrase in config.
 */
#if SSLEAY_VERSION_NUMBER >= 0x00904000L
static int ssl_pem_password_cb(char *buf, int size, int rwflag, void *userdata)
#else
static int ssl_pem_password_cb(char *buf, int size, int rwflag)
#endif
{
  if(ssl_pem_passphrase)
    {
      strncpy(buf, ssl_pem_passphrase, size - 1);
      buf[size - 1] = '\0';
  
      return strlen(buf);
    }
  else      
    log(L_WARN, "Passphrase required but not in config file (S-Line)");
  
  return -1;
}

/*
 * generate a temporary RSA key
 */
static RSA *ssl_tmp_rsa_cb(SSL *ssl, int x, int len)
{
  log(L_DEBUG, "Creating a temporary RSA key of length %i...", len);

  /* ... */
  
  return NULL;
}

/* 
 * seed prng from 'filename', when 'filename' doesn't exist from
 * '/dev/urandom'....
 * 
 * TODO: seed from another prng (gettimeofday() stuff) when '/dev/urandom'
 *       doesn't exist.
 */
int ssl_seed_prng(char *filename)
{
  struct stat st;
  int readbytes, writebytes;
  
  if(stat(filename, &st))
    readbytes = RAND_load_file("/dev/urandom", RANDOM_SIZE);
  else
    readbytes = RAND_load_file(filename, RANDOM_SIZE);
  
  if(readbytes <= 0)
    {
      log(L_WARN, "Unable to retrieve any random data from %s", filename);
      return 0;
    }
  else
    log(L_DEBUG, "Read %d random bytes from %s", readbytes, filename);
 
  writebytes = RAND_write_file(filename);
  
  if(writebytes <= 0)
    {
      log(L_WARN, "Failed to write random data to %s: %s", filename, strerror(errno));
      return 0;
    }
  else
    log(L_DEBUG, "Wrote %d random bytes to %s", writebytes, filename);
  
  return readbytes;  
}

/* 
 * initialize diffie-hellman key exchange algorithm.
 * (currently unused)
 */
int init_diffie_hellman(SSL_CTX *ctx)
{
  DH *dh;
  BIO *bio;
  
  if(!(bio = BIO_new_file(CEPATH, "r")))
    return -1;

  if(!(dh = PEM_read_bio_DHparams(bio, NULL, NULL
#if SSLEAY_VERSION_NUMBER >= 0x00904000L
                                 , NULL
#endif
                                 )))
    {
      BIO_free(bio);
      return -1;
    }
  
  SSL_CTX_set_tmp_dh(ctx, dh);
  log(L_NOTICE, "Diffie-Hellman initialized with %u bit key", DH_size(dh) * 8);
  BIO_free(bio);    
  DH_free(dh);
  
  return 0;
}

/*
 * do handshake on an outgoing connection.
 */
int ssl_connect(struct Client *cptr)
{
  int ssl_err;

  cptr->ssl = SSL_new(client_ctx);
  CHK_NULL(cptr->ssl);    
  SSL_set_connect_state(cptr->ssl);
  SSL_set_fd(cptr->ssl, cptr->fd);
  
  SSL_set_app_data(cptr->ssl, (char *)cptr);
  
  if((ssl_err = SSL_connect(cptr->ssl)) <= 0)
    {
      ssl_err = SSL_get_error(cptr->ssl, ssl_err);

      if(ssl_err == SSL_ERROR_WANT_WRITE)
        {
          errno = EAGAIN;
          return 0;
        }
      
      return ssl_error(ssl_err, SSL_CONNECT, cptr);
    }

//  cptr->ssl_isconnecting = 0;
  sendto_realops("SSL_connect is done");
  return 0;
}

/*
 * do handshake on an incoming connection.
 */
int ssl_accept(struct Client *cptr, int fd)
{
  int ssl_err;

  SSL_set_app_data(cptr->ssl, (char *)cptr);

  if((ssl_err = SSL_accept(cptr->ssl)) <= 0)
    {
      switch(ssl_err = SSL_get_error(cptr->ssl, ssl_err))
        {
          case SSL_ERROR_SSL:
            if(errno == 0)
              return -1;
          case SSL_ERROR_SYSCALL:
            errno = EINTR;
            return -1;
          case SSL_ERROR_WANT_READ:
          case SSL_ERROR_WANT_WRITE:
            errno = EAGAIN;
            return -1;
          default:
            return ssl_error(ssl_err, SSL_ACCEPT, cptr);
        }
      /* NOTREACHED */
      return -1;
    }

  sendto_realops_lev(L_DEBUG, "SSL handshake done: %p", cptr->ssl);

  return 0;
}

/*
 * read from SSL
 */
int ssl_read(struct Client *cptr, void *buf, size_t sz)
{
  int len, ssl_err;

  do
    {      
      cptr->read_blocked_on_write = 0;
      cptr->read_blocked = 0;
      
      len = SSL_read(cptr->ssl, buf, sz);
      
      switch(ssl_err = SSL_get_error(cptr->ssl, len))
        {
          case SSL_ERROR_NONE:
            break;
          case SSL_ERROR_ZERO_RETURN:
            return 0;
          case SSL_ERROR_WANT_READ:
            cptr->read_blocked = 1;
            errno = EWOULDBLOCK;
            return -1;
          case SSL_ERROR_WANT_WRITE:
            cptr->read_blocked_on_write=1;
            break;
          case SSL_ERROR_SSL:
          case SSL_ERROR_SYSCALL:
            if(errno == EWOULDBLOCK || errno == EAGAIN ||
               errno == EINTR)
            return -1;
            break;
          default:
            return ssl_error(ssl_err, SSL_READ, cptr);
        }      
    }
  while(SSL_pending(cptr->ssl) && !cptr->read_blocked);
  
  return len;
}
  

/*
 * write to SSL
 */
int ssl_write(struct Client *cptr, const void *buf, size_t sz)
{
  int len, ssl_err;

  cptr->write_blocked_on_read = 0;
  
  len = SSL_write(cptr->ssl, buf, sz);
  
  switch(ssl_err = SSL_get_error(cptr->ssl, len))
    {
      case SSL_ERROR_NONE:
        break;
      case SSL_ERROR_ZERO_RETURN:
        return 0;
      case SSL_ERROR_WANT_READ:
        cptr->write_blocked_on_read = 1;
        break;
      case SSL_ERROR_WANT_WRITE:
        cptr->write_blocked = 1;
        errno = EWOULDBLOCK;
        return -1;
      case SSL_ERROR_SSL:
      case SSL_ERROR_SYSCALL:
        if(errno == EWOULDBLOCK || errno == EAGAIN ||
           errno == EINTR)
          return -1;
        break;
      default:
        return ssl_error(ssl_err, SSL_ACCEPT, cptr);
    }
  
  return len;
}

/*
 * ssl socket cleanup
 */
int ssl_shutdown(SSL *ssl)
{
  char i;
  int rc;

  rc = 0;
  for(i = 0; i < 4; i++)
    if((rc = SSL_shutdown(ssl)))
      break;

  return rc;
}

/*
 * raise an error
 */
static int ssl_error(int error, int where, struct Client *sptr)
{
  /* don`t alter errno */
  int errtmp = errno;
  char *errstr = strerror(errtmp);
  char *ssl_errstr, *ssl_func;

  switch(where)
    {
      case SSL_READ:
        ssl_func = "SSL_read()";
        break;
      case SSL_WRITE:
        ssl_func = "SSL_write()";
        break;
      case SSL_ACCEPT:
        ssl_func = "SSL_accept()";
        break;
      case SSL_CONNECT:
        ssl_func = "SSL_connect()";
        break;
      default:
        ssl_func = "undefined SSL func [this is a bug]";
    }

    switch(error)
      {
        case SSL_ERROR_NONE:
          ssl_errstr = "No error";
          break;
        case SSL_ERROR_SSL:
          ssl_errstr = "Internal OpenSSL error or protocol error";
          break;
        case SSL_ERROR_WANT_READ:
          ssl_errstr = "OpenSSL functions requested a read()";
          break;
        case SSL_ERROR_WANT_WRITE:
          ssl_errstr = "OpenSSL functions requested a write()";
          break;
        case SSL_ERROR_WANT_X509_LOOKUP:
          ssl_errstr = "OpenSSL requested a X509 lookup which didn`t arrive";
          break;
        case SSL_ERROR_SYSCALL:
          ssl_errstr = "Underlying syscall error";
          break;
        case SSL_ERROR_ZERO_RETURN:
          ssl_errstr = "Underlying socket operation returned zero";
          break;
        case SSL_ERROR_WANT_CONNECT:
          ssl_errstr = "OpenSSL functions wanted a connect()";
          break;
        default:
          ssl_errstr = "Unknown OpenSSL error (huh?)";
    }

  if(error != SSL_ERROR_ZERO_RETURN)
    {
      sendto_realops("%s to "
                     "%s!%s@%s aborted with%serror (%s). [%s]",
                     ssl_func, *sptr->name ? sptr->name : "<unknown>",
                     (sptr->user && sptr->username) ? sptr->username : "<unregistered>", sptr->sockhost,
                     (errno > 0) ? " " : " no ", errstr, ssl_errstr);
      log(L_WARN, "SSL error in %s: %s [%s]", ssl_func, errstr, ssl_errstr);
    }
  else
    errtmp = ECONNRESET;
  
  /* if we reply() something here, we might just trigger another
   * ssl_error() call and loop until a stack overflow...
   * the client won`t get the ERROR : ... string, but this is
   * the only way to do it.

   * IRC protocol wasn`t SSL enabled .. --vejeta
   */

  errno = errtmp ? errtmp : ECONNABORTED; /* Software caused connection abort */
  sptr->flags |= FLAGS_DEADSOCKET;
  return -1;
}


/* 
 * server certificate callback
 */
static int ssl_server_verify(int preverify_ok, X509_STORE_CTX *x509_ctx)
{
  sendto_realops_flags(FLAGS_DEBUG, "ssl_server_verify callback (preverify: %u)", preverify_ok);
  return ssl_verify(preverify_ok, x509_ctx, 1, ssl_svl);
}

/*
 * client certificate callback
 */
static int ssl_client_verify(int preverify_ok, X509_STORE_CTX *x509_ctx)
{
  sendto_realops_flags(FLAGS_DEBUG, "ssl_client_verify callback (preverify: %u)", preverify_ok);
  return ssl_verify(preverify_ok, x509_ctx, 0, ssl_cvl);
}

/*
 * generic certificate callback
 */
static int ssl_verify(int preverify_ok, X509_STORE_CTX *x509_ctx, int isserver, int level)
{
  char           buf[256];
//  X509          *err_cert;
  int            err, depth;
  SSL           *ssl;
  struct Client *cptr;
  X509_OBJECT    ret;
  
  X509_NAME_oneline(X509_get_subject_name(x509_ctx->current_cert),
                    buf, 255);
  
  if(level == 0)
    {
      sendto_realops_flags(FLAGS_DEBUG, "ignoring %s verify callback: cert = '%s'",
                           isserver ? "server" : "client", buf);
      return 1; /* accept connection */
    }
  
//  err_cert = X509_STORE_CTX_get_current_cert(x509_ctx);
  err = X509_STORE_CTX_get_error(x509_ctx);
  depth = X509_STORE_CTX_get_error_depth(x509_ctx);
  
  /*
   * Retrieve the pointer to the SSL of the connection
   * currently treated and the cptr from the object.
   */
  
  ssl = X509_STORE_CTX_get_ex_data(x509_ctx, SSL_get_ex_data_X509_STORE_CTX_idx());
  cptr = (struct Client *)SSL_get_app_data(ssl);
  
/*  X509_NAME_oneline(X509_get_subject_name(err_cert), buf, 256);*/
  cptr->verify_ok = -1;
  
  if(!preverify_ok)
    {
      sendto_realops_flags(FLAGS_DEBUG, "verify error: depth = %u, error = %u, cert = '%s'",
                           depth, err, buf);
      return 0; /* reject connection */
    }
  
  if(level == 3 && depth == 0 && 
     X509_STORE_get_by_subject(x509_ctx, X509_LU_X509,
     X509_get_subject_name(x509_ctx->current_cert), &ret) != 1)
    {
      sendto_realops_flags(FLAGS_DEBUG, "verify error: no cert for %s", buf);
      return 0; /* reject connection */
    }
  
  cptr->verify_ok = 1;
  sendto_realops_flags(FLAGS_DEBUG, "verify ok: depth = %u, cert = '%s'", depth, buf);
  
  return preverify_ok;
}

/*
 * send certificate info as notice to a client
 */
int ssl_show_cert(struct Client *sptr, struct Client *acptr, X509 *cert)
{
  ASN1_UTCTIME *utc;
  time_t from, to, current = time(NULL);
  struct tm *tm;
  char v_from[32], v_to[32];
  char buf[256];
  long serial, version;
  ASN1_INTEGER *asn1_i;

  /* ugly time string formatting */
  utc = X509_get_notBefore(cert);
  from = ASN1_UTCTIME_get(utc);
  tm = localtime(&from);
  strftime(v_from, 32, "%F %H:%M", tm);
  utc = X509_get_notAfter(cert);
  to = ASN1_UTCTIME_get(utc);
  tm = localtime(&to);
  strftime(v_to, 32, "%F %H:%M", tm);
      
  sendto_one(sptr, ":%s NOTICE %s :*** Certificate information about %s",
             me.name, sptr->name, acptr->name);

  asn1_i = X509_get_serialNumber(cert);
  serial = ASN1_INTEGER_get(asn1_i);
  version = X509_get_version(cert);
  sendto_one(sptr, ":%s NOTICE %s :Serial #%u (X509 version %u)", me.name, sptr->name, serial, version);
  
  X509_NAME_oneline(X509_get_subject_name(cert), buf, 255);
  sendto_one(sptr, ":%s NOTICE %s :Subject: %s", me.name, sptr->name, buf);
  
  X509_NAME_oneline(X509_get_issuer_name(cert), buf, 255);
  sendto_one(sptr, ":%s NOTICE %s :Issuer:  %s", me.name, sptr->name, buf);
  
  if(acptr->verify_ok)
    sendto_one(sptr, ":%s NOTICE %s :Verify:  %s", me.name, sptr->name,
               (acptr->verify_ok == 1) ? "Ok" : "Failed");
  
  sendto_one(sptr, ":%s NOTICE %s :Valid from %s to %s (%s)",
             me.name, sptr->name, v_from, v_to,
             (from <= current && current < to) ? "valid" : 
             (from > current) ? "not yet valid" : "expired");
      
  return 0;
}

/*
 * send info about an ssl session
 */ 
int ssl_show_info(struct Client *sptr, struct Client *acptr)
{
  SSL_CIPHER *cipher;
  const char *cipher_name;
  int cipher_bits;
  time_t t, to;
  struct tm *tm;
  char timebuf[32];
  
  if(!acptr->ssl)
    {
      log(L_WARN, "ssl_show_info called but non-ssl link (this is a bug)");
      return 0;
    }
  
  if(!(cipher = SSL_get_current_cipher(acptr->ssl)))
    {
      log(L_WARN, "could not retrieve cipher for %s", acptr->name);
      return 0;
    }

  cipher_name = SSL_CIPHER_get_name(cipher);
  SSL_CIPHER_get_bits(cipher, &cipher_bits);
       
  sendto_one(sptr, ":%s NOTICE %s :*** Session information about link between %s and %s",
             me.name, sptr->name, acptr->name, me.name);
  
  sendto_one(sptr, ":%s NOTICE %s :Protocol:    %s", 
             me.name, sptr->name, SSL_CIPHER_get_version(cipher));
  sendto_one(sptr, ":%s NOTICE %s :Cipher:      %s (%i bits)",
             me.name, sptr->name, cipher_name, cipher_bits);
  
  if(acptr->ssl->session)
    {
      t = SSL_SESSION_get_time(acptr->ssl->session);
      to = SSL_SESSION_get_timeout(acptr->ssl->session);
  
      tm = localtime(&t);
      strftime(timebuf, 32, "%F %H:%M", tm);      
      
      sendto_one(sptr, ":%s NOTICE %s :Established: %s", 
                 me.name, sptr->name, timebuf);
      
      sendto_one(sptr, ":%s NOTICE %s :Timeout:     %d seconds",
                 me.name, sptr->name, to);
    }
  
  return 1;
}

/*
 * m_cert()
 *
 * parv[0] = sender prefix
 * parv[1] = server or client
 *
 * show certificate info about clients.
 */
int     m_cert(struct Client *cptr,
                struct Client *sptr,
                int parc,
                char *parv[])
{  
  struct Client *acptr, *a2cptr = &me;
  X509 *cert;
#ifdef PACE_CERT
  static time_t last_used = 0L;
#endif
  
  /* only allow this command from ssl links */
  if(!IsSSL(cptr) || !cptr->ssl)
    {
      if(MyClient(sptr))
        sendto_one(cptr, form_str(ERR_UNKNOWNCOMMAND),
                   me.name, sptr->name, "CERT");
      return 0;
    }
  
#ifdef PACE_CERT  
  if(MyClient(sptr) && !IsAnOper(sptr))
    {
      if((last_used + PACE_WAIT) > CurrentTime)
        {
          sendto_one(sptr,form_str(RPL_LOAD2HI), me.name,parv[0]);
          return 0;
        }
      else
        last_used = CurrentTime;
    }
#endif
  
  /*
   * find the server/client to
   * show certificate info from.
   */
  if(parc <= 1)
    {
      /* set acptr to &me when local client and no arg */
      if(MyClient(sptr))
        acptr = &me;
      else
        return 0;
    }
  else
    if(valid_hostname(parv[1]))
      {
        /* arg looks like a hostname, find server struct */
        if((acptr = find_server(parv[1])) == NULL)
          {
            sendto_one(sptr, form_str(ERR_NOSUCHSERVER), me.name, sptr->name, parv[1]);
            return 0;
          }
        
        a2cptr = acptr;
      }
    else
      {      
        /* could be a nickname, find client struct */
        if((acptr = find_chasing(sptr, parv[1], NULL)) == NULL)
          return 0;
        
        a2cptr = acptr->servptr;
      }
  
  /* check wether client/server is in ssl mode */
  if((IsPerson(acptr) && !IsSSLUmode(acptr)) ||
     (IsServer(acptr) && !IsSSLServ(acptr)))    
    {
      sendto_one(sptr, form_str(RPL_NOTSSL), me.name, sptr->name);
      return 0;
    }
  
  /*
   * send out the request when it is not a
   * local connect and not &me.
   * else dump information.
   */
  
  if(MyConnect(acptr) || &me == acptr)
    {
      if(&me == acptr)
        cert = SSL_get_certificate(cptr->ssl);
      else
        cert = SSL_get_peer_certificate(acptr->ssl);
      
      if(cert == NULL)
        sendto_one(sptr, ":%s NOTICE %s :*** %s has no certificate",
                   me.name, sptr->name, acptr->name);
      else
        ssl_show_cert(sptr, acptr, cert);
      
      if(acptr == &me)
        ssl_show_info(sptr, sptr);
      else
        ssl_show_info(sptr, acptr);
      
      return 1;     
    }
  else
    /* send out the stuff */      
    sendto_one(a2cptr, ":%s CERT %s", sptr->name, acptr->name);
  
  return 1;
}

#endif
